import Platform from './lib/platform.js'
import config from './config.js'
import cm from './lib/codemirror.js'
window.cm = cm

function Editor() {
  this.open = async () => {
    if(config.statusBarAtTop)
      document.body.appendChild(statusbar.element)

    console.log(cm, cm.EditorView)

    this.view = new cm.EditorView({
      state: await openToDocument()
        .catch(console.error),
      parent: document.body
    })

    if(!config.statusBarAtTop)
      document.body.appendChild(statusbar.element)

    this.doc.setView(this.view)
    this.view.focus()
  }

  // this.updateFilename = () => {
  //   this.doc.path = Platform.dirname(this.doc.path) + '/' + statusbar.filename.value
  // }

  const openToDocument = async () => {
    let doc = new Doc(config.landingDocument.render(), config.landingDocument.name)

    this.doc = doc
    statusbar.filename.value = config.landingDocument.name
    return await doc.createState()
      .catch(err => console.error("Could not open landing document", err))
  }

  this.doc = null
}

function Statusbar() {
  this.selectFilename = () => {
    lastFilename = this.filename.value

    this.filename.value = editor.doc.getPath()
    this.filename.select()
    this.filename.selectionStart = editor.doc.getPath().lastIndexOf('/') + 1
  }

  this.element = document.createElement('nav')
  this.filename = document.createElement('input')
  let lastFilename = ''

  this.element.appendChild(this.filename)

  this.filename.addEventListener('click', this.selectFilename)

  this.filename.addEventListener('keydown', (event) => {
    switch(event.key) {
      case 'Enter':
      case 'Escape':
        event.preventDefault()
        editor.view.focus()
        break
    }
  })

  this.filename.addEventListener('focusout', event => {
    if(this.filename.value == '') {
      this.filename.value = lastFilename
    } else {
      let path = this.filename.value

      editor.doc.setPath(path)
      this.filename.value = Platform.filename(path)
    }
  })
}

function Doc(content, initialPath = '') {
  // if(path) {
  //   let workspaces = Platform.storage.get('workspaces')
  // }

  this.createState = async () => {
    let extensions = [
      Doc.zen.of([
        cm.lineNumbers(),
        cm.highlightActiveLine(),
        cm.highlightActiveLineGutter(),
        cm.highlightSpecialChars(),
      ]),
      cm.foldGutter(),
      cm.drawSelection(),
      cm.dropCursor(),
      cm.EditorState.allowMultipleSelections.of(true),
      cm.indentOnInput(),
      cm.syntaxHighlighting(cm.defaultHighlightStyle, { fallback: true }),
      cm.bracketMatching(),
      cm.closeBrackets(),
      cm.autocompletion(),
      cm.rectangularSelection(),
      cm.crosshairCursor(),
      cm.highlightSelectionMatches(),
      Keymaps.default
    ]

    if(this.language)
      extensions.push(
        await langManager.resolveExtension(this.language)
          .then(extension => Doc.language.of(extension))
      )

    return cm.EditorState.create({
      extensions,
      doc: content
    })
  }

  this.setPath = newPath => {
    path = newPath
    this.setLanguage()
  }

  this.getPath = () =>
    path

  this.setView = (newView) => {
    view = newView
    this.setLanguage()
  }

  this.setLanguage = (newLanguage = getLanguage()) => {
    let oldLanguage = this.language
    this.language = newLanguage

    if(!view)
      return

    if(this.language == null) {
      setViewLanguageExtension([])
    } else if(oldLanguage == null || oldLanguage.name !== this.language.name) {
      langManager.resolveExtension(this.language)
        .then(setViewLanguageExtension)
        .catch(err => console.error('Failed to set language extension', err))
    }
  }

  const getLanguage = () => {
    let ext = Platform.ext(path)

    if(ext) {
      return langManager.exts[ext]
    }
  }

  const setViewLanguageExtension = (extension) => {
    view.dispatch({
      effects: Doc.language.reconfigure(extension)
    })
  }

  let path = initialPath
  let view = null
  this.language = getLanguage()
}

Doc.open = async path => {
  return new Doc() 
}

Doc.tabSize = new cm.Compartment
Doc.zen = new cm.Compartment
Doc.language = new cm.Compartment

const Keymaps = {
  default: cm.keymap.of([
    ...cm.closeBracketsKeymap,
    ...cm.defaultKeymap,
    ...cm.searchKeymap,
    ...cm.foldKeymap,
    ...cm.completionKeymap,
    ...cm.lintKeymap,
    cm.indentWithTab,
    {
      key: 'Ctrl-r',
      run({ state }, event) {
        event.preventDefault()
        let mainSelection = state.selection.main

        if(!mainSelection.empty)
          statusbar.filename.value = state.sliceDoc(mainSelection.from, mainSelection.to)
        statusbar.selectFilename()
      }
    }
  ])
}

function LanguageManager() {
  this.resolveExtension = (language) =>
    import(language.import)
      .then(mod => language.createExtension(mod.default))

  this.exts = {}

  for(let language of config.languages) {
    for(let ext of language.exts)
      this.exts[ext] = language
  }
}

window.addEventListener('load', () => {
  window.statusbar = new Statusbar()
  window.editor = new Editor()
  window.langManager = new LanguageManager()
  editor.open()
})